/*
 * Execution Context
 *
 * Copyright (C) 2009-2011 Udo Steinberg <udo@hypervisor.org>
 * Economic rights: Technische Universitaet Dresden (Germany)
 *
 * This file is part of the NOVA microhypervisor.
 *
 * NOVA is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * NOVA is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License version 2 for more details.
 */

#include "bits.h"
#include "ec.h"
#include "assert.h"
#include "cpu.h"
#include "ptab.h"
#include "multiboot.h"
#include "elf.h"
#include "selectors.h"

Ec * Ec::current;

// used for idle() and root_invoke()
Ec::Ec (void (*f)(), mword mbi) : cont (f)
{
   regs.eax = mbi;		// Store MultiBoot Info to user-space eax register
   regs.cs  = SEL_USER_CODE;
   regs.ds  = SEL_USER_DATA;
   regs.es  = SEL_USER_DATA;
   regs.ss  = SEL_USER_DATA;
   regs.efl = 0x200;           // IF = 1
}

// only used by syscall create thread (EC+SC)
Ec::Ec (mword eip)
{
   cont = ret_user_iret;
   regs.cs  = SEL_USER_CODE;
   regs.ds  = SEL_USER_DATA;
   regs.es  = SEL_USER_DATA;
   regs.ss  = SEL_USER_DATA;
   regs.efl = 0x200;           // IF = 1
   regs.eip = eip;
}

void Ec::ret_user_sysexit()
{
   asm volatile ("lea %0, %%esp;"
   "popa;"
   "sti;"
   "sysexit"
   : : "m" (current->regs) : "memory");
   
   UNREACHED;
}

void Ec::ret_user_iret()
{
   asm volatile ("lea %0, %%esp;"
   "popa;"
   "pop %%gs;"
   "pop %%fs;"
   "pop %%es;"
   "pop %%ds;"
   "add $8, %%esp;"
   "iret"
   : : "m" (current->regs) : "memory");
   
   UNREACHED;
}

void Ec::idle()
{
   for (;;)
      
      UNREACHED;
}

void Ec::root_invoke()
{
   // Bootloader passed us the "Multiboot information" data
   // structure. Among other things, it contains the physical address
   // of the user space application we want to execute.
   
   // To access the data at physical address, we have to "map" them
   // to virtual memory. Ptab::remap() creates a temporary mapping of
   // one page.
   Multiboot * mbi = static_cast<Multiboot *>(Ptab::remap (current->regs.eax));
   
   if (!(mbi->flags & 8) || (mbi->mods_count != 1))
      panic ("Exactly ONE multi boot module is required.\n");
   
   // Load module desciptor
   Multiboot_module mod = *static_cast<Multiboot_module *>(Ptab::remap (mbi->mods_addr));
   
   printf ("Application physical location %#x - %#x (%u bytes)\n",
           mod.mod_start, mod.mod_end, mod.mod_end - mod.mod_start);
   
   // Setup virtual memory mapping for user space application
   
   // Allocate one page of memory for user space stack
   void *stack = Kalloc::allocator.alloc_page(1);
   
   // Map the stack page at address 0x1000
   Ptab::insert_mapping (1 * PAGE_SIZE, Kalloc::virt2phys (stack), Ptab::PRESENT | Ptab::RW | Ptab::USER);
   
   // Check ELF header existence
   if(sizeof(Eh) >= mod.mod_end - mod.mod_start){
      panic("ELF error\n");
   };
   
   // Access the ELF file
   Eh* elf = (Eh*) Ptab::remap(mod.mod_start);
   
   // Check ELF magic
   if(elf->ei_magic != 0x464c457f){// 7f 'E' 'L' 'F', reversed byte-order
      panic("ELF error\n");
   };
   
   // Check program header
   if(elf->ph_size != sizeof(Ph) || elf->ph_count < 1 || elf->ph_offset < sizeof(Eh) || elf->ph_offset + elf->ph_size * elf->ph_count > mod.mod_end - mod.mod_start){
      panic("ELF error\n");
   };
   
   // Read program headers
   Ph* phs = (Ph*) ((mword) elf + elf->ph_offset);
   
   for(unsigned int i = 0; i < elf->ph_count; i++){
      if(phs[i].type == Ph::PT_LOAD){
         // Check length
         if(phs[i].f_size < 1 || phs[i].f_size != phs[i].m_size || phs[i].f_offs + phs[i].f_size > mod.mod_end - mod.mod_start){
            panic("ELF error\n");
         };
         
         // Map pages
         mword ph_begin = align_dn(phs[i].v_addr, PAGE_SIZE);
         mword j_count = (align_up(phs[i].v_addr + phs[i].m_size, PAGE_SIZE) - ph_begin) / PAGE_SIZE;
         for(mword j = 0; j < j_count; j++){
            Ptab::insert_mapping(ph_begin + j*PAGE_SIZE, mod.mod_start + phs[i].f_offs + j*PAGE_SIZE, Ptab::PRESENT | Ptab::USER | ((phs[i].flags & Ph::PF_W) ? Ptab::RW : 0));
         };
      };
   };
   
   // Jump to user-space to entry point
   current->regs.eip = elf->entry;
   current->regs.cs = SEL_USER_CODE;
   current->regs.efl = 0x200;
   current->regs.esp = 0x2000;
   current->regs.ss = SEL_USER_DATA;
   
   Ec::ret_user_iret();
   
   UNREACHED;
}

void Ec::handle_tss()
{
   panic ("Task gate invoked\n");
}

/**
 * Returns whether specified page is already mapped
 * @param virt - virtual address to check mapping for
 * @return true/false - present/not-present
 */
static bool present(mword virt){
   mword* pdir = static_cast<mword*>(Kalloc::phys2virt(Cpu::cr3()));
   if((pdir[virt >> 22] & Ptab::PRESENT) == 0){
      return false;
   };
   
   mword* ptab = static_cast<mword*>(Kalloc::phys2virt (pdir[virt >> 22] & ~PAGE_MASK));
   
   return ((ptab[(virt >> PAGE_BITS) & 0x3ff] & Ptab::PRESENT) != 0);
} // Ptab::present

/**
 * Allocates additional memory for user program
 * @param where - position in program's virtual memory to map new memory into
 * @param size - size of the memory in bytes
 * @return where on success, NULL otherwise
 */
static mword syscall_alloc_mem(mword where, unsigned size){
   // check virtual pointer validity
   if(where < 0x1000 || size < 1){
      return NULL;
   };
   
   // check presence of mapped pages affected by memory request
   mword begin = align_dn(where, PAGE_SIZE);
   mword count = (align_up(where + size, PAGE_SIZE) - begin) / PAGE_SIZE;
   for(unsigned int i = 0; i < count; i++){
      // check user/kernel space bounds
      if(begin+i*PAGE_SIZE >= 0xC0000000){
         return NULL;
      };
      if(present(begin + i*PAGE_SIZE)){
         return NULL;
      };
   };
   
   // map new pages
   mword pages = Kalloc::virt2phys(Kalloc::allocator.alloc_page(count));
   for(unsigned int i = 0; i < count; i++){
      Ptab::insert_mapping(begin + i*PAGE_SIZE, pages + i*PAGE_SIZE, Ptab::PRESENT | Ptab::USER | Ptab::RW);
   };
   
   return where;
} // Ec::syscall_alloc_mem

void Ec::syscall_handler (uint8 number)
{
   Sys_regs * r = current->sys_regs();
   
   switch (number) {
      
      case 1: { // write_to_terminal
         char *data = reinterpret_cast<char*>(r->esi);
         for (unsigned i = 0; i < r->edi; i++)
            printf("%c", data[i]);
         break;
      }
      
      case 2:
         r->eax = syscall_alloc_mem(r->esi, r->edi);
         break;
      default:
         printf ("unknown syscall %d\n", number);
         break;
   };
   
   ret_user_sysexit();
   
   UNREACHED;
}

bool Ec::handle_exc_ts (Exc_regs *r)
{
   if (r->user())
      return false;
   
   // SYSENTER with EFLAGS.NT=1 and IRET faulted
   r->efl &= ~0x4000; // nested task eflag
   
   return true;
}

static const char *vec_desc[] = {
   "Divide error",
   "RESERVED",
   "NMI Interrupt",
   "Breakpoint",
   "Overflow",
   "BOUND Range Exceeded",
   "Undefined Opcode",
   "No Math Coprocessor",
   "Double Fault",
   "Coprocessor Segment Overrun",
   "Invalid TSS",
   "Segment Not Present",
   "Stack-Segment Fault",
   "General Protection",
   "Page Fault",
};


void Ec::handle_exc (Exc_regs *r)
{
   printf ("Ec::handle_exc %s (eip=%#lx cr2=%#lx)\n", vec_desc[r->vec], r->eip, r->cr2);
   printf ("eax=%#lx ebx=%#lx ecx=%#lx edx==%#lx\n", r->eax, r->ebx, r->ecx, r->edx);
   printf ("esi=%#lx edi=%#lx ebp=%#lx esp==%#lx\n", r->esi, r->edi, r->ebp, r->esp);
   
   if (r->vec == Cpu::EXC_TS && handle_exc_ts (r))
      return;
   
   panic ("unhandled kernel exception\n");
}
